/*
 * Odin Zapte/Dahdi Driver for n64 streaming
 *
 * Based on previous works by Mark Spencer <markster@digium.com>
 * Based on previous works, designs, and archetectures conceived and
 * written by Jim Dixon <jim@lambdatel.com>.
 *
 * Copyright (C) 
 *
 * All rights reserved.
 *
 */

/*
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2 as published by the
 * Free Software Foundation. See the LICENSE file included with
 * this program for more details.
 */

#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <dahdi/kernel.h>
#include <linux/moduleparam.h>
#include <linux/delay.h> 
#include <linux/jiffies.h>
#include <linux/proc_fs.h>

#include <linux/miscdevice.h>
#include <asm/uaccess.h>
#include <linux/fs.h>
#include <linux/time.h>
#include <linux/kthread.h>
#include <linux/completion.h>

#include <linux/wait.h>
#include <linux/smp_lock.h>
#include <linux/suspend.h>
#include <linux/slab.h>

#include <linux/netdevice.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)

#include <linux/freezer.h>
#include <linux/syscalls.h>

#ifndef PF_FREEZE
#define PF_FREEZE TIF_FREEZE
#endif

#endif

#define NEW_DAHDI_API

//#define VERBOSE

#ifdef __arm
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#endif

#define ETH_P_TDMOP 0x5555


PACKED struct VirtualSpanParamsT {
  const char             *hwDev;
  const char             *dstMac;
  short                  nChannel; // 0, 1, 2, 3
  short                  nType;    // 1 - T1, 2 - E1
};

#define	TYPE_T1	1		/* is a T1 card */
#define	TYPE_E1	2		/* is an E1 card */

#define MAX_CARDS      1
#define MAX_SPANS      32
#define SPANS_PER_CARD 32
#define TIMESLOTS      32

static int cospans = 1; // Spans count to be initialized at startup. MAX_SPANS;

//static int cnt_rx = 0;
//static int cnt_tx = 0;

static int cnt_eth_rx = 0;
static int cnt_eth_tx = 0;
static int cnt_eth_rx_bytes = 0;
static int cnt_eth_tx_bytes = 0;
static int cnt_n64_rx = 0;
//static int cnt_n64_tx = 0;
static int cnt_rx_broken = 0;
static int cnt_rx_frames = 0;
//static int cnt_tx_frames = 0;
static int cnt_rx_wrong_source = 0;
static int cnt_n64_rx_bytes_broken = 0;
static int cnt_n64_rx_frames_broken = 0;
static int cnt_n64_rx_seq_broken = 0;
        
static int cnt_rx_bytes_in_frames_in_last_sec = 0;
static unsigned long cnt_last_jiffies = 0;
static int cnt_rx_bytes_per_sec;

#define WITH_PROCFS

#ifdef WITH_PROCFS
static struct proc_dir_entry* otx_proc_dir;
static struct proc_dir_entry* otx_proc_dir_spans;
static struct proc_dir_entry* otx_proc_rxdata;
static struct proc_dir_entry* otx_proc_stats;
static struct proc_dir_entry* otx_proc_cospans;
#endif

#pragma pack(push,1)
struct __attribute__ ((packed)) N64Header {
    unsigned char  cVer;
    unsigned char  cOffset;
    unsigned short nCh;
    unsigned long  nSeq;
    unsigned short coFrames;
    unsigned short coTs;
};
#pragma pack(pop)


// Uncomment the define below to show the runtime debug info..
#define OTXDAHDI_DEBUG_RUNTIME


PACKED struct otx_chan {
	/* Private pointer for channel.  We want to know our
	   channel and span */
	struct otxcard *card;
	int span;	/* Index from 0 */
};


PACKED struct otx_span {
	/* Private pointer for span.  We want to know our
	   span number and pointer to the otx card device */
	struct otxcard *card;
	int span;	/* Index from 0 */
#ifdef NEW_DAHDI_API
	struct dahdi_span dahdi_span;
#endif
};


 /* This structure exists one per card */
PACKED struct otxcard {
  int num;									// Which card we are 
  int syncs     [SPANS_PER_CARD];						// sync sources 
  int psyncs    [SPANS_PER_CARD];						// span-relative sync sources 
  int alarmtimer[SPANS_PER_CARD];						// Alarm timer 
  char *type;									// Type of card 
  int order;									// Order 
  int flags;									// Device flags 
  struct dahdi_span spans[SPANS_PER_CARD];					// Spans
#ifdef NEW_DAHDI_API 
  struct otx_span otx_spans[SPANS_PER_CARD];					// Span data 
#else
  struct otx_span tspans[SPANS_PER_CARD];					// Span data 
#endif
  struct dahdi_chan **chans[SPANS_PER_CARD];					// Pointers to blocks of 24(30/31) contiguous dahdi_chans for each span 
  struct otx_chan tchans[SPANS_PER_CARD * TIMESLOTS];				// Channel user data 
  unsigned char txsigs[SPANS_PER_CARD][16];					// Copy of tx sig registers 
  int loopupcnt[SPANS_PER_CARD];						// loop up code counter 
  int loopdowncnt[SPANS_PER_CARD];						// loop down code counter 
  int spansstarted;								// number of spans started 
  unsigned char ec_chunk1[SPANS_PER_CARD][TIMESLOTS][DAHDI_CHUNKSIZE];		// first EC chunk buffer 
  unsigned char ec_chunk2[SPANS_PER_CARD][TIMESLOTS][DAHDI_CHUNKSIZE];		// second EC chunk buffer
  int cardtype;									// card type, T1 or E1 
  unsigned int *datxlt;								// pointer to datxlt structure 
  int bOpen;
  unsigned char readchunks[SPANS_PER_CARD*DAHDI_CHUNKSIZE*TIMESLOTS];
  unsigned char writechunks[SPANS_PER_CARD*DAHDI_CHUNKSIZE*TIMESLOTS];
  unsigned char nSrcCh[SPANS_PER_CARD*TIMESLOTS];				// Dacs timeslots
  unsigned char dacs_buffer[SPANS_PER_CARD*DAHDI_CHUNKSIZE*TIMESLOTS];
}*cards[MAX_CARDS];

PACKED struct otxvirtualspan { 
  short			nSpan;
  short                 nChannel;
  struct net_device     *hwDev;
  char                  dstMac[ETH_ALEN];
  short                 nType;
  short                 bEnabled;
  int                   duhdiRxBufPos;
  int                   duhdiTxBufPos;
  unsigned long         nRxSeq;  
  unsigned long         nTxSeq;  
  struct proc_dir_entry* proc;
}virtualspans[MAX_SPANS];

/* un-define this if you dont want NON-REV A hardware support */
/* #define	NONREVA 1 */

#define	SYNCSELF 0
#define	SYNC1	1
#define	SYNC2	2
#define	SYNC3	3
#define	SYNC4	4
#define SYNCEXTERN 5

/* signalling bits */
#define	OTX_ABIT 8
#define	OTX_BBIT 4

static int cardtype = 2;
//static int highestorder;

#if 1
/* translations of data channels for 24 channels in a 32 bit PCM highway */
unsigned datxlt_t1[] = { 
    1 ,2 ,3 ,5 ,6 ,7 ,9 ,10,11,13,14,15,17,18,19,21,22,23,25,26,27,29,30,31 };

/* translations of data channels for 30/31 channels in a 32 bit PCM highway */
unsigned datxlt_e1[] = { 
    1 ,2 ,3 ,4 ,5 ,6 ,7 ,8 ,9 ,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
	25,26,27,28,29,30,31 };
#endif

PACKED struct transmit_command {
  int span;
  int len;
  char mac[ETH_ALEN];
  struct N64Header hdr;
};

static DECLARE_MUTEX(transmit_sem);
static DEFINE_SPINLOCK(transmit_command_lock);

#define transmit_COMMAND_LIST_COUNT 512
static struct transmit_command g_transmit_command_list[transmit_COMMAND_LIST_COUNT];
static struct transmit_command* g_p_transmit_command_list_end = (struct transmit_command*) ((void*)g_transmit_command_list + sizeof(struct transmit_command)*transmit_COMMAND_LIST_COUNT);
static struct transmit_command* g_p_transmit_command_read = g_transmit_command_list;
static struct transmit_command* g_p_transmit_command_write = g_transmit_command_list;


static DECLARE_WAIT_QUEUE_HEAD(transmit_wait);
static DECLARE_COMPLETION(transmit_exited);
static int transmit_pid = 0;

#define transmit_MAXIMUM_PACKET_SIZE 1472 + sizeof(struct N64Header)
static char g_buftransmitPacket[transmit_MAXIMUM_PACKET_SIZE];


static int otx_add_virtual_span(short nSpan, struct VirtualSpanParamsT params);
static int otx_remove_virtual_span(short nSpan);
static int otx_dahdi_register_span(struct otxcard *card, int nSpan);
static int otx_dahdi_unregister_span(struct otxcard *card, int nSpan);
static int otx_create_dahdi_span(struct otxcard* card, int nSpan);
static int otx_free_dahdi_span(struct otxcard *card, int nSpan);

#ifdef WITH_PROCFS

ssize_t odin_module_read_rxdata(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{
  int i = 0;
  for (i = 0; i < size && i < SPANS_PER_CARD*DAHDI_CHUNKSIZE*TIMESLOTS; i++) {
    buffer[i] = cards[0]->readchunks[i];
  }
  return i;
}

ssize_t odin_module_read_stats(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{
  int i = 0;
  i = snprintf( buffer, size, "Ehernet:\n"
                              "  RX Packets: %d\n"
                              "  RX Bytes: %d\n"
                              "  RX broken packets: %d\n"
                              "  RX wrong source packets: %d\n"
                              "  TX Packets: %d\n"
                              "  TX Bytes: %d\n\n"
                              "N64 Protocol:\n"
                              "  RX packets: %d\n"
                              "  RX frames: %d\n"
                              "  RX BPS: %d\n"
                              "  RX broken bytes: %d\n"
                              "  RX broken frames: %d\n"
                              "  RX broken sequences: %d\n",
                              cnt_eth_rx, cnt_eth_rx_bytes, cnt_rx_broken, cnt_rx_wrong_source, 
                              cnt_eth_tx, cnt_eth_tx_bytes,
                              cnt_n64_rx, cnt_rx_frames, cnt_rx_bytes_per_sec, cnt_n64_rx_bytes_broken, cnt_n64_rx_frames_broken, cnt_n64_rx_seq_broken);
  return i;
}
        
ssize_t odin_module_read_spancfg(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{            
    struct otxvirtualspan* pSpan = (struct otxvirtualspan*) data;
    int i = 0;
    
    if (!pSpan) {
      i = snprintf( buffer, size, "Wrong span handle: %p\n", data);      
      return i;
    }
    
    i = snprintf( buffer, size, "Channel: %d\n"
				"Dev: %s\n"
				"Remote MAC: %02X:%02X:%02X:%02X:%02X:%02X\n"
				"Enabled: %s\n"
				"RX Seq: %lu\n"
				"TX Seq: %lu\n\n"
				"To configure span write parameters to this file discriptor.\n"
				"Supported parameters:\n"
				"  mac    - MAC address of remote side\n"
				"  if     - Ethernet interface\n"
				"  ch     - Channel number\n"
				"  enable - [1/0] Enable or disable span\n\n"
				"Example: echo \"mac=FF:FF:FF:FF:FF:FF if=eth0 ch=%d enable=1\" > /proc/otxdahdin64/spans/%d\n",
				pSpan->nChannel, pSpan->hwDev ?  pSpan->hwDev->name : "Error",
				(unsigned char)pSpan->dstMac[0], (unsigned char)pSpan->dstMac[1], (unsigned char)pSpan->dstMac[2], (unsigned char)pSpan->dstMac[3], (unsigned char)pSpan->dstMac[4], (unsigned char)pSpan->dstMac[5],
				pSpan->bEnabled ? "Yes" : "No" , pSpan->nRxSeq, pSpan->nTxSeq, pSpan->nChannel, pSpan->nSpan);

    
    return i;
}

ssize_t odin_module_write_spancfg(struct file* file, const char* buffer, unsigned long count, void *data)
{
  struct otxvirtualspan* pSpan = (struct otxvirtualspan*) data;
  unsigned short mac[6];
  char strIfc[64];
  int nCh;  
  int bEnable;
  const char* p = buffer;
  const char* pend = buffer+count;
    
  if (!pSpan) {
    return 0;
  }
  
  while (p < pend) {
    while (p < pend && (*p == ' ' || *p == '\r' || *p == '\n')) { p++; }
    if (sscanf( p, "mac=%2hx:%2hx:%2hx:%2hx:%2hx:%2hx", &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]) == 6) {
      pSpan->dstMac[0] = mac[0] & 0xFF;
      pSpan->dstMac[1] = mac[1] & 0xFF;    
      pSpan->dstMac[2] = mac[2] & 0xFF;    
      pSpan->dstMac[3] = mac[3] & 0xFF;    
      pSpan->dstMac[4] = mac[4] & 0xFF;    
      pSpan->dstMac[5] = mac[5] & 0xFF;    
    }
  
    if (sscanf( p, "if=%64s", strIfc) == 1) {
      pSpan->hwDev = dev_get_by_name(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
				    &init_net,
#endif
    				    strIfc); 
      if (!pSpan->hwDev) {
        printk("Wrong Ethernet interface '%s'. Use default: eth0", strIfc);
        pSpan->hwDev = dev_get_by_name(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
					&init_net,
#endif
    					"eth0");
      }
    }
  
    if (sscanf( p, "ch=%d", &nCh) == 1) {
      pSpan->nChannel = nCh;      
    }                            

    if (sscanf( p, "enable=%d", &bEnable) == 1) {
      pSpan->bEnabled = bEnable ? 1 : 0;      
    }  
        
    while (p < pend && *p != ' ' && *p != '\r' && *p != '\n') { p++; }
  }
                  
  return count;
}
                          
                          
ssize_t odin_module_read_cospans(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{            
    int i = 0;
    
    i = snprintf( buffer, size, "%d", cospans);
    
    return i;
}

ssize_t odin_module_write_cospans(struct file* file, const char* buffer, unsigned long count, void *data)
{
    int newCospans, x;
    struct otxcard *card = cards[0];
    
    if (sscanf( buffer, "%d", &newCospans) != 1) {
      return count;
    }
    
    if (newCospans >= 1 && newCospans <= MAX_SPANS) {
	if (cospans > newCospans) {
	    // Delete spans
	    for (x = newCospans; x < cospans; x++) {
		otx_dahdi_unregister_span(card, x);
		otx_remove_virtual_span(x);
		otx_free_dahdi_span(card, x);
	    }
	    cospans = newCospans;
	} else if (newCospans > cospans) {
	    // Add spans
	    for (x = cospans; x < newCospans; x++) {
		struct VirtualSpanParamsT vspan;
		vspan.hwDev = "eth0";
                vspan.dstMac = "\xFF\xFF\xFF\xFF\xFF\xFF";
                vspan.nChannel = x;
                vspan.nType = cardtype;
                                        
                otx_create_dahdi_span(card, x);                                                  
                otx_add_virtual_span(x, vspan);                                                            
        	otx_dahdi_register_span(card, x);		
	    }
	    cospans = newCospans;	    
	}
    } else {
	printk("otxdahdin64: Wrong new spans count. Must be from 1 to 32\n");
    }
    
    return count;
}

                          
#endif

static int chan_num(struct dahdi_chan *chan)
{
  int i;
  for (i = 0; i < SPANS_PER_CARD; i++) {
    if (chan->span == cards[0]->spans + i) {
      return TIMESLOTS * i + chan->chanpos;
    }
  }
  return 0xff;
}

static int otx_dacs(struct dahdi_chan *dst, struct dahdi_chan *src)
{
  int rc = 0;

  return rc;
   
} 

static ssize_t span_read(struct file * file, char * buf, size_t count, loff_t *ppos)
{
  return 0;
}

static ssize_t span_write(struct file *file, const char *buf, size_t size, loff_t *ppos)
{
  return 0;
}
 
static struct file_operations span_fops = {
  .owner    = THIS_MODULE,
  .read    = span_read,
  .write   = span_write,
};

static struct miscdevice span_devs[SPANS_PER_CARD] = {
    { MISC_DYNAMIC_MINOR, "odin_span0", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span1", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span2", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span3", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span4", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span5", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span6", &span_fops },
    { MISC_DYNAMIC_MINOR, "odin_span7", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span8", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span9", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span10", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span11", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span12", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span13", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span14", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span15", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span16", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span17", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span18", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span19", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span20", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span21", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span22", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span23", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span24", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span25", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span26", &span_fops },    
    { MISC_DYNAMIC_MINOR, "odin_span27", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span28", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span29", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span30", &span_fops },        
    { MISC_DYNAMIC_MINOR, "odin_span31", &span_fops }
  };


static int otx_span_shutdown(struct dahdi_span *span)
{
	int tspan;
#ifdef NEW_DAHDI_API 
    struct otx_span *const p = container_of(span, struct otx_span, dahdi_span);
#else
	struct otx_span *p = span->pvt;
#endif
	tspan = p->span + 1;
	if (tspan < 0) {
		printk("otxdahdi: Span '%d' isn't us?\n", span->spanno);
		return -1;
	}

	//spans_rx_config &= ~(1 << (span->spanno-1));
	//spans_tx_config &= ~(1 << (span->spanno-1));

#ifdef VERBOSE
	printk("otxdahdi: Shutting span %d down\n", span->spanno);
#endif

	return 0;
}


static int otx_span_startup(struct dahdi_span *span)
{
	int tspan;
#ifdef NEW_DAHDI_API 
    struct otx_span *const p = container_of(span, struct otx_span, dahdi_span);
#else
	struct otx_span *p = span->pvt;
#endif
	tspan = p->span + 1;
	if (tspan < 0) {
		printk("otxdahdi: Span '%d' isn't us?\n", span->spanno);
		return -1;
	}

#ifdef VERBOSE
	printk("otxdahdi: Starting span %d up\n", span->spanno);
#endif
	return 0;
}


static int otx_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
#if 0
	switch(cmd) {
	default:
		return -ENOTTY;
	}
#endif
	return 0;
}


static int otx_span_config(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
#ifdef VERBOSE
	printk("otxdahdi: Configuring span %d\n", span->spanno);
#endif

	if (span->flags & DAHDI_FLAG_RUNNING)
		return otx_span_startup(span);

	return 0;
}

static int otx_span_changeconfig(struct dahdi_chan *chan, int sigtype)
{
#ifdef VERBOSE
	struct otx_chan *p = chan->pvt;

	printk("otxdahdi: Configuring channel %d (%s) sigtype %d card_type=%d\n", 
		chan->channo, chan->name, sigtype, p->card->cardtype);
#endif
	
	return 0;
}


static int otx_chan_open(struct dahdi_chan *chan)
{
#ifdef VERBOSE
	printk("otxdahdi: Openning channel %d\n", chan_num(chan));
#endif

	return 0;
}


static int otx_chan_close(struct dahdi_chan *chan)
{
#ifdef VERBOSE
	printk("otxdahdi: Closing channel %d\n", chan_num(chan));
#endif
	return 0;
}

#ifdef NEW_DAHDI_API 
static const struct dahdi_span_ops otx_span_ops = {
	.owner = THIS_MODULE,
	.spanconfig = otx_span_config,
	.chanconfig = otx_span_changeconfig,
	.startup = otx_span_startup,
	.shutdown = otx_span_shutdown,
	.dacs = otx_dacs,
	.rbsbits = 0,
	.maint = 0,
	.open = otx_chan_open,
	.close = otx_chan_close,
	.ioctl = otx_ioctl,
};
#endif

static int otx_add_virtual_span(short nSpan, struct VirtualSpanParamsT params) {

        struct otxcard *card = cards[0];
        int x, y, c;
        char tmpstr[32];
        
        struct otxvirtualspan* pSpan = 0;

        if (nSpan >= MAX_SPANS) { 
          return 0;
        }
        
        pSpan = (struct otxvirtualspan*) &(virtualspans[nSpan]);
        
        pSpan->nSpan = nSpan;
        pSpan->bEnabled = 1;
        pSpan->duhdiRxBufPos = 0;
        pSpan->duhdiTxBufPos = 0;        
	memcpy(pSpan->dstMac, params.dstMac, ETH_ALEN);
	pSpan->hwDev = dev_get_by_name(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
					&init_net,
#endif
					params.hwDev
	);
	pSpan->nType = params.nType;       
	pSpan->nChannel = params.nChannel;       
	pSpan->nTxSeq = 0;
	pSpan->nRxSeq = 0;
	
#ifdef WITH_PROCFS	
	sprintf(tmpstr, "%d", nSpan);
	pSpan->proc = create_proc_entry( tmpstr, 0444, otx_proc_dir_spans );
        if( pSpan->proc != NULL ) {
//    	    pSpan->proc->owner = THIS_MODULE;
    	    pSpan->proc->data = (void*)pSpan;
	    pSpan->proc->read_proc = odin_module_read_spancfg;
	    pSpan->proc->write_proc = odin_module_write_spancfg;	    
	} else {
	  printk("Failed create proc entry for span %d.\n", nSpan);
	}
#endif	                                                                                
	
      	x = nSpan;

	sprintf(card->spans[x].name, "Otx/%d/%d", card->num, x + 1);
	snprintf(card->spans[x].desc, sizeof(card->spans[x].desc) - 1,
		"Remote Alvis %s Board %d Span %d",
		(params.nType == TYPE_T1)  ?  "T1"  :  "E1", card->num, x + 1);
	card->spans[x].manufacturer = "OdinTS";
	dahdi_copy_string(card->spans[x].devicetype, "otxcard", sizeof(card->spans[x].devicetype));
	snprintf(card->spans[x].location, sizeof(card->spans[x].location) - 1,"n/a");
#ifndef NEW_DAHDI_API
	card->spans[x].spanconfig = otx_span_config;
	card->spans[x].chanconfig = otx_span_changeconfig;
	card->spans[x].startup = otx_span_startup;
	card->spans[x].shutdown = otx_span_shutdown;
	card->spans[x].dacs = otx_dacs;
	card->spans[x].rbsbits = 0;
	card->spans[x].maint = 0;
	card->spans[x].open = otx_chan_open;
	card->spans[x].close  = otx_chan_close;
#endif
	if (params.nType == TYPE_T1) {
		card->spans[x].channels = 24;
		card->spans[x].deflaw = DAHDI_LAW_MULAW;
		card->spans[x].linecompat = DAHDI_CONFIG_AMI | DAHDI_CONFIG_B8ZS 
					 | DAHDI_CONFIG_D4 | DAHDI_CONFIG_ESF;
		card->spans[x].spantype = "T1";
	} else {
		card->spans[x].channels = 31;
		card->spans[x].deflaw = DAHDI_LAW_ALAW;
		card->spans[x].linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS 
				 	 | DAHDI_CONFIG_CRC4;
		card->spans[x].spantype = "E1";
	}
	card->spans[x].chans = card->chans[x];
	card->spans[x].flags = DAHDI_FLAG_RBS;
#ifdef NEW_DAHDI_API 
	card->spans[x].ops = &otx_span_ops;
	card->otx_spans[x].card = card;
	card->otx_spans[x].span = x;
#else
	card->spans[x].ioctl = otx_ioctl;
	card->spans[x].pvt = &card->tspans[x];
	card->tspans[x].card = card;
	card->tspans[x].span = x;
#endif	
	init_waitqueue_head(&card->spans[x].maintq);
	for (y=0;y<card->spans[x].channels;y++) {
		struct dahdi_chan *mychans = card->chans[x][y];
		sprintf(mychans->name, "Otx/%d/%d/%d", card->num, x + 1, y + 1);
		mychans->sigcap = DAHDI_SIG_CLEAR | DAHDI_SIG_SF | DAHDI_SIG_DACS;
		c = (x * card->spans[x].channels) + y;
		mychans->pvt = &card->tchans[c];
		mychans->chanpos = y + 1;
		card->tchans[c].span = x;
		card->tchans[c].card = card;
	}
        return 1;
}

static int otx_remove_virtual_span(short nSpan) {
    struct otxvirtualspan* pSpan = 0;
    char tmp[32];

    if (nSpan >= MAX_SPANS) {
	return 0;
    }
                          
    pSpan = (struct otxvirtualspan*) &(virtualspans[nSpan]);
    memset(pSpan, 0, sizeof(struct otxvirtualspan));
    
#ifdef WITH_PROCFS
    sprintf(tmp, "%d", nSpan);
    remove_proc_entry( tmp, otx_proc_dir_spans );                        
#endif
    
    return 0;                                
}

static int otx_create_dahdi_span(struct otxcard* card, int nSpan) 
{
    int x = nSpan, f = 0;
    card->chans[x] = kmalloc(sizeof(*card->chans[x]) * 31,GFP_USER);
    if (!card->chans[x]){
	printk(KERN_ERR "otxdahdin64: Fail to allocate bytes for  struct dahdi_chan \n");
	return -ENOMEM;
    }
    memset(card->chans[x],0,sizeof(*card->chans[x]) * 31);
    for (f = 0; f < 31; f++) {
	if (!(card->chans[x][f] = kmalloc(sizeof(*card->chans[x][f]), GFP_USER))) {
	    printk(KERN_ERR "otxdahdin64: Not enough memory for chans[%d][%d]\n", x, f);
	    return -ENOMEM;
	}
	memset(card->chans[x][f], 0, sizeof(*card->chans[x][f]));
	card->chans[x][f]->readchunk = &(card->readchunks[x * 32 * DAHDI_CHUNKSIZE + (f+1)*DAHDI_CHUNKSIZE]);
	card->chans[x][f]->writechunk = &(card->writechunks[x * 32 * DAHDI_CHUNKSIZE + (f+1)*DAHDI_CHUNKSIZE]);
    }
    return 0;	
}

static int __devinit odin_module_launch_card(struct otxcard *card)
{
#ifdef VERBOSE
	printk("otxdahdi: odin_module_launch_card \n");
#endif
	if (card->spans[0].flags & DAHDI_FLAG_REGISTERED) {
		return 0;
	}

#ifdef VERBOSE
	printk("otxdahdi: Launching card: %d\n", card->order);
#endif
#if 0
	for (x = 0; x < cospans; x++) {
            if (!virtualspans[x].bEnabled) { 
              continue;
            }
#ifdef VERBOSE
    	    printk("otxdahdi: Register span %s\n", card->spans[x].name);
#endif
	
	    if (dahdi_register(&card->spans[x], 0)) {
		printk(KERN_ERR "Unable to register span %s\n", card->spans[x].name);
		for (x--;x >= 0; x--) {
		    dahdi_unregister(&card->spans[x]);
		}
		return -1;
	    }
	}
#endif		
	card->bOpen = 1;
	
	return 0;
}

static int otx_dahdi_register_span(struct otxcard *card, int nSpan)
{
#ifdef VERBOSE
    printk("otxdahdi: Register span %s\n", card->spans[nSpan].name);
#endif
    if (dahdi_register(&card->spans[nSpan], 0)) {
	printk(KERN_ERR "Unable to register span %s\n", card->spans[nSpan].name);
	return -1;
    }
    if ( misc_register(&(span_devs[nSpan])) )printk(KERN_ERR"Unable to register span%d misc device\n", nSpan);
    return 0;
}

static int otx_dahdi_unregister_span(struct otxcard *card, int nSpan)
{
#ifdef VERBOSE
    printk("otxdahdi: Unregister span %s\n", card->spans[nSpan].name);
#endif
    if(card->spans[nSpan].flags & DAHDI_FLAG_REGISTERED) {
	dahdi_unregister(&card->spans[nSpan]);
    }
    
    misc_deregister(&(span_devs[nSpan]));
                          
    return 0;
}

static int otx_free_dahdi_span(struct otxcard *card, int nSpan) 
{
    unsigned int f;
    if (!card->chans[nSpan]) {
	return 0;
    }
    for (f = 0; f < (card->cardtype == TYPE_E1 ? 31 : 24); f++) {
	if (card->chans[nSpan][f]) {
		kfree(card->chans[nSpan][f]);
		card->chans[nSpan][f] = 0;
	}
    }

    kfree(card->chans[nSpan]);
    card->chans[nSpan] = 0;
    
    return 0;
}


static void free_otxcard(struct otxcard *card)
{
    unsigned int x;
    for (x = 0; x < cospans; x++) {
	otx_free_dahdi_span(card, x);
    }
    kfree(card);
    card = 0;
}

int transmit_queue(int span, const char* mac, int len, struct N64Header hdr)
{
        unsigned long flags;
        struct transmit_command *cmd;

        spin_lock_irqsave(&transmit_command_lock, flags);

        cmd = g_p_transmit_command_write;
        cmd->span = span;
        memcpy(cmd->mac, mac, ETH_ALEN);
        cmd->len  = len;
        cmd->hdr  = hdr;

        g_p_transmit_command_write++;
        if (g_p_transmit_command_write >= g_p_transmit_command_list_end) {
          g_p_transmit_command_write = g_transmit_command_list;
        }

        wake_up(&transmit_wait);

        spin_unlock_irqrestore(&transmit_command_lock, flags);
        return 1;
}

static int transmit_get_command(struct transmit_command* cmd)
{
        unsigned long flags;

        spin_lock_irqsave(&transmit_command_lock, flags);

        if (g_p_transmit_command_read == g_p_transmit_command_write) {
            spin_unlock_irqrestore(&transmit_command_lock, flags);
            return 0;
        }

        memcpy(cmd, g_p_transmit_command_read, sizeof(struct transmit_command));

        g_p_transmit_command_read++;
        if (g_p_transmit_command_read >= g_p_transmit_command_list_end) {
          g_p_transmit_command_read = g_transmit_command_list;
        }

        spin_unlock_irqrestore(&transmit_command_lock, flags);

        return 1;
}

static int otxdahdi_send_packet(int span, char* mac, const char* msg, int size) {  
        struct sk_buff *skb;
        struct net_device *dev;
//        unsigned char* header;
//        struct N64Header* hdr = (struct N64Header*) msg;

        struct otxvirtualspan* pSpan = 0;
        
        if (span >= MAX_SPANS) {
    	    return 0;
        }
                          
        
        pSpan = &(virtualspans[span]);

        dev = pSpan->hwDev;
        
        if (!dev) {
          pSpan->hwDev = dev = dev_get_by_name(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
						&init_net,
#endif
        					"eth0"
          );
          return 0;
        }
        
        if (!dev) {
          return 0;
        }

        skb = dev_alloc_skb(size + sizeof(struct N64Header) + dev->hard_header_len);
        if (!skb) {
          return -ENOMEM;
        }
        skb_reserve(skb, dev->hard_header_len);

        memcpy(skb_put(skb,size), msg, size);

        skb->protocol = __constant_htons(ETH_P_TDMOP);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
        skb_set_network_header(skb,0);
#else
        skb->nh.raw = skb->data;
#endif
        skb->dev = dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
        dev_hard_header(skb, dev, ETH_P_TDMOP, mac, dev->dev_addr, skb->len);
#else
        if (dev->hard_header)
          dev->hard_header(skb, dev, ETH_P_TDMOP, mac, dev->dev_addr, skb->len);
#endif
        cnt_eth_tx++;
        cnt_eth_tx_bytes += size;
#if 1
        dev_queue_xmit(skb);
#else
        kfree_skb(skb);
#endif
        return 0;
}

static int otxdahdi_transmit_process_span(int span, char* mac, int len, struct N64Header* hdr) {
        struct N64Header* transmitHdr = (struct N64Header*) g_buftransmitPacket;
        char* data = g_buftransmitPacket + hdr->cOffset;
        char* p = data;
        char* pend = g_buftransmitPacket + transmit_MAXIMUM_PACKET_SIZE;
        unsigned short coFrames = ntohs(hdr->coFrames);
        unsigned short coTs = ntohs(hdr->coTs);        
        int nTs = 0, nFrame = 0, j = 0;
	struct otxvirtualspan* pSpan = 0;
	
	if (span >= MAX_SPANS) {
	  return 0;
	}
	pSpan = &(virtualspans[span]);        
	
        memcpy(transmitHdr, hdr, sizeof(struct N64Header));
        transmitHdr->nSeq = htonl(pSpan->nTxSeq++);
	
        if (len > transmit_MAXIMUM_PACKET_SIZE) {
          return 0;
        }
        
        j = pSpan->duhdiTxBufPos;
                
//        dahdi_transmit( cards[0]->spans + span );

        for (nFrame = 0; nFrame < coFrames && p < pend; ++nFrame) {
          if (j == DAHDI_CHUNKSIZE) {
            dahdi_transmit( cards[0]->spans + span );
            j = 0;
          }
          for (nTs = 0; nTs < coTs; ++nTs) {
            *p++ = cards[0]->writechunks[((span * TIMESLOTS) + nTs) * DAHDI_CHUNKSIZE + j];
            if (p >= pend) {
              break;
            }
          }
          ++j;
        }
        
        pSpan->duhdiTxBufPos = j;

        otxdahdi_send_packet(span, mac, g_buftransmitPacket, len); 

        return 1;
}

static void transmit_handle_commands(void)
{
        struct transmit_command cmd;

        while ((transmit_get_command(&cmd))) {
          otxdahdi_transmit_process_span(cmd.span, cmd.mac, cmd.len, &cmd.hdr);
        }
}

static int transmit_thread(void *nothing)
{
        lock_kernel();
        daemonize("ktransmitn64");
        allow_signal(SIGTERM);

        do {
            transmit_handle_commands();
            wait_event_interruptible(transmit_wait, g_p_transmit_command_read != g_p_transmit_command_write);
            if (current->flags & PF_FREEZE)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
                    refrigerator();
#else            
                    refrigerator(PF_FREEZE);
#endif                    
        } while (!signal_pending(current));

        printk(KERN_DEBUG "otxdahdin64: ktransmitn64 exiting\n");

        unlock_kernel();
        complete_and_exit(&transmit_exited, 0);
        return 0;
}


static int otxdahdi_receive_process_span(int span, const char* mac, char* buf, size_t len) {
        struct N64Header* hdr = (struct N64Header*) buf;
        char* data = buf + hdr->cOffset;
        char* p = data;
        char* pend = buf + len;
        unsigned short coFrames = ntohs(hdr->coFrames);
        unsigned short coTs = ntohs(hdr->coTs);
        unsigned long nSeq = ntohl(hdr->nSeq);
        int nTs = 0, nFrame = 0, j = 0;
        struct otxvirtualspan* pSpan = &(virtualspans[span]);        
        
        if (pSpan->nRxSeq) {
          if (nSeq != pSpan->nRxSeq + 1) {
            cnt_n64_rx_seq_broken++;
          }
        }
        pSpan->nRxSeq = nSeq;
        
        cnt_n64_rx++;        
        j = pSpan->duhdiRxBufPos;
 
        for (nFrame = 0; nFrame < coFrames && p < pend; ++nFrame) {
          if (j == DAHDI_CHUNKSIZE) {
            dahdi_receive( cards[0]->spans + span);  
            j = 0;
          }
          for (nTs = 0; nTs < coTs; ++nTs) {
            cards[0]->readchunks[((span * TIMESLOTS) + nTs) * DAHDI_CHUNKSIZE + j] = *p++;
            if (p >= pend) {
              break;
            }
          }
          ++j;
          cnt_rx_frames++;          
        }
        if (j == DAHDI_CHUNKSIZE) {
          dahdi_receive( cards[0]->spans + span);  
          j = 0;
        }
        
        pSpan->duhdiRxBufPos = j;
        
        cnt_n64_rx_bytes_broken += pend - p;
        cnt_n64_rx_frames_broken += coFrames - nFrame;
        
        cnt_rx_bytes_in_frames_in_last_sec += p - data;
        
        if (jiffies_to_msecs(jiffies) - cnt_last_jiffies >= 2000) {
          cnt_last_jiffies = jiffies_to_msecs(jiffies);
          cnt_rx_bytes_per_sec = cnt_rx_bytes_in_frames_in_last_sec/2;
          cnt_rx_bytes_in_frames_in_last_sec = 0;
        }        
 
        transmit_queue(span, mac, len, *hdr);

        return 1;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
static int otxdahdi_receive_packet(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt, 
                                struct net_device *orig_dev)
#else
static int otxdahdi_receive_packet(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt)
#endif
{
        int x;
        struct N64Header* hdr = 0;
        int nSpan = -1;
        int nRxCh = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
        if (skb_is_nonlinear(skb) && skb_linearize(skb) != 0) {
#else
        if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0) {
#endif
          kfree_skb(skb);
          cnt_rx_broken++;
          return 0;
        }

        cnt_eth_rx++;
        cnt_eth_rx_bytes += skb->len;


        if (skb->len >= sizeof(struct N64Header)) {
          hdr = (struct N64Header*) skb->data;
        } else {
          cnt_rx_broken++;
          kfree_skb(skb);
          return 0;
        }
        
        nRxCh = ntohs(hdr->nCh);

        for (x = 0; x < MAX_SPANS; x++) {
          if ((virtualspans[x].bEnabled == 1) && (virtualspans[x].nChannel == nRxCh)) {
            if ((memcmp(virtualspans[x].dstMac, "\xFF\xFF\xFF\xFF\xFF\xFF", ETH_ALEN) == 0) ||
        	(memcmp(virtualspans[x].dstMac, 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)            
		skb_mac_header(skb)
#else
                skb->mac.raw
#endif                
                 + ETH_ALEN, ETH_ALEN) == 0))
             {
               nSpan = x;
               break;
             } else {
                kfree_skb(skb);
                cnt_rx_wrong_source++;
                return 0;              
            }
          }
        }

        if ((nSpan == -1) || (nSpan >= MAX_SPANS)) {
          cnt_rx_broken++;
          kfree_skb(skb);
          return 0;
        }        

        otxdahdi_receive_process_span(nSpan, 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)            
		skb_mac_header(skb)
#else
                skb->mac.raw
#endif                        
				 + ETH_ALEN, skb->data, skb->len);  

        kfree_skb(skb);
        return 0;
}


static struct packet_type OtxDahdiPacketType = {
	.type = __constant_htons(ETH_P_TDMOP),
	.dev = NULL,
	.func = otxdahdi_receive_packet,
};


// Called when module load 
static int __init odin_module_start(void) {
	int res = 0;
	int x;
	struct otxcard *card;
	
	if (cospans > MAX_SPANS) {
	  cospans = MAX_SPANS;
	}

	card = kmalloc(sizeof(struct otxcard), GFP_USER);
	if (!card){
		printk(KERN_ERR"otxdahdi: odin_module_start() Fail to allocate bytes for  struct otxcard \n");
		return -ENOMEM;
	}

	memset(card,0,sizeof(struct otxcard));
	memset(card->nSrcCh, 0xff, SPANS_PER_CARD*TIMESLOTS);

        dev_add_pack(&OtxDahdiPacketType);

	for (x = 0; x < MAX_CARDS; x++) {
		if (!cards[x]) break;
	}
	if (x >= MAX_CARDS) {
		printk(KERN_ERR"No cards[] slot available!!\n");
		goto err_out_release_all;
	}
	card->num = x;
	cards[x] = card;
#ifdef VERBOSE
	printk("odin_module_start(): TRY to Launch cards as appropriate\n");
#endif
	card->cardtype = cardtype;
	card->datxlt = datxlt_e1;

#ifdef WITH_PROCFS
	otx_proc_dir = proc_mkdir( "otxdahdin64", NULL );
	if( otx_proc_dir == NULL ) {
		printk( "otxdahdi: Failed to create proc directory\n" );
	} else {
//		otx_proc_dir->owner = THIS_MODULE;
		otx_proc_rxdata = create_proc_entry( "rxdata", 0444, otx_proc_dir );
		if( otx_proc_rxdata != NULL ) {
//			otx_proc_rxdata->owner = THIS_MODULE;
			otx_proc_rxdata->read_proc = odin_module_read_rxdata;
		}				
		otx_proc_stats = create_proc_entry( "stats", 0444, otx_proc_dir );
		if( otx_proc_stats != NULL ) {
//			otx_proc_stats->owner = THIS_MODULE;
			otx_proc_stats->read_proc = odin_module_read_stats;
		}	
		otx_proc_dir_spans = proc_mkdir( "spans", otx_proc_dir);		
		otx_proc_cospans = create_proc_entry( "cospans", 0444, otx_proc_dir );
		if( otx_proc_cospans != NULL ) {
			otx_proc_cospans->read_proc = odin_module_read_cospans;
			otx_proc_cospans->write_proc = odin_module_write_cospans;			
		}	
		
	}
#endif

        
	/* Launch cards as appropriate */
#ifdef VERBOSE
	printk("odin_module_start(): Launch cards as appropriate\n");
#endif
#if 0
	x = 0;
	for (;;) {
		/* Find a card to activate */
		f = 0;
		for (x=0;x < MAX_CARDS;x++) {
			if (cards[x]->order <= highestorder) {
				odin_module_launch_card(cards[x]);
				if (cards[x]->order == highestorder)
					f = 1;
			}
		}
		/* If we found at least one, increment the highest order and search again, otherwise stop */
		if (f) 
			highestorder++;
		else
			break;
	}
#else
	odin_module_launch_card(card);
#endif

        memset(virtualspans, 0, sizeof(virtualspans));

        for (x = 0; x < cospans; x++) {
          struct VirtualSpanParamsT vspan;
          vspan.hwDev = "eth0";          
          vspan.dstMac = "\xFF\xFF\xFF\xFF\xFF\xFF";
          vspan.nChannel = x;
          vspan.nType = cardtype;
          
          otx_create_dahdi_span(card, x);
          
          otx_add_virtual_span(x, vspan);
          
          otx_dahdi_register_span(card, x);
        }


        if (!(transmit_pid = kernel_thread(transmit_thread, NULL, CLONE_KERNEL))) {
          printk(KERN_WARNING "otxdahdi_n64: Failed to start kttransmitn64 thread\n");
        }


//        for (x = 0; x < cospans; x++) {
//          if ( misc_register(&(span_devs[x])) )printk(KERN_ERR"Unable to register span%d misc device\n", x);
//        }
        
	return res;

err_out_release_all:
	if (card) {
		free_otxcard( card );
	}
	if( res == 0 )
		return -ENODEV;
	else
		return res;
}

// Called when module unload 
static void __exit odin_module_stop(void)
{

	struct otxcard *card = cards[ 0 ];
	int x;
	
        dev_remove_pack(&OtxDahdiPacketType);

#ifdef VERBOSE
        printk("odin_module_stop(): Unregistering spans...\n");
#endif
        
        for (x = 0; x < cospans; x++) {
    	    otx_dahdi_unregister_span(card, x);
    	    otx_remove_virtual_span(x);
  	  //if(card->spans[x].flags & DAHDI_FLAG_REGISTERED) {
	  //	dahdi_unregister(&card->spans[x]);
	  //} 
	}
#ifdef VERBOSE
        printk("odin_module_stop(): free otxcard structure...\n");
#endif

	if(card){
		free_otxcard(card);
	}

#ifdef WITH_PROCFS
        for (x = 0; x < cospans; x++) {
          char tmp[32];
          sprintf(tmp, "%d", x);
          remove_proc_entry( tmp, otx_proc_dir_spans );	
        }
        remove_proc_entry( "spans", otx_proc_dir );
	remove_proc_entry( "rxdata", otx_proc_dir );	
	remove_proc_entry( "stats", otx_proc_dir );		
	remove_proc_entry( "cospans", otx_proc_dir );			
	remove_proc_entry( "otxdahdin64", NULL );	
#endif

        if (transmit_pid) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)            
	    struct pid * pid = find_vpid(transmit_pid);
	    kill_pid(pid, SIGTERM, 1);	    
//	    struct task_struct* ts = find_task_by_vpid(transmit_pid);
//	    send_sig(SIGTERM, ts, 1);
#else            
	    kill_proc(transmit_pid, SIGTERM, 1);
#endif	    
            wait_for_completion(&transmit_exited);
        }

	printk("Unregistered otxdahdi\n");
}


MODULE_AUTHOR("Odin TeleSystems Inc");
MODULE_DESCRIPTION("OTX DAHDI N64 Network Driver");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

module_init(odin_module_start);
module_exit(odin_module_stop);

module_param(cardtype, int, 0600);
MODULE_PARM_DESC(cardtype, "If 1 is a T1 card, If 2 is an E1 card");

module_param(cospans, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(cospans, "Spans count to be registered in DAHDI module");

